#include <Arduino.h>
#include <ArduinoOTA.h>
#include <ESP8266WiFi.h>

#include <i2s.h>
#include <i2s_reg.h>

#include <WebSocketsServer.h>

#define ESP_WIFI_SSID "OpenWrt"
#define ESP_WIFI_PSK "3JGJXC3R"

#define ESP_OTA_HOSTNANE "" // hostname (defaults to esp8266-[ChipID])
#define ESP_OTA_PASSWORD "" // no authentication by default

/* WI-FI
 * ========================================================================= */
void ESP_Boot_WiFiConnect(const char *ssid, const char *psk)
{
    WiFi.mode(WIFI_STA);
    WiFi.begin(ESP_WIFI_SSID, ESP_WIFI_PSK);
    
    Serial.print("[WIFI] Connecting..");
    for (int i = 0; WiFi.status() != WL_CONNECTED; i++) {
        Serial.print(".");
        digitalWrite(LED_BUILTIN, i % 2);
        delay(250);
    }
    Serial.println();
    digitalWrite(LED_BUILTIN, HIGH); // Turn LED off
    
    Serial.printf("[WIFI] Connected to %s (IP: %s)\n", WiFi.SSID().c_str(), WiFi.localIP().toString().c_str());
}

/* OTA
 * ========================================================================= */
void ESP_Boot_EnableOTA(const char *hostname, const char* password)
{
    ArduinoOTA.setHostname(ESP_OTA_HOSTNANE);        
    ArduinoOTA.setPassword(ESP_OTA_PASSWORD);
    
    ArduinoOTA.onStart([]() {
        Serial.printf("[OTA] Start updating %s\n", (ArduinoOTA.getCommand() == U_FLASH ? "sketch" : "filesystem"));
    });
    ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
        Serial.printf("[OTA] Progress: %u%%\r", (progress / (total / 100)));
    });
    ArduinoOTA.onEnd([]() {
        Serial.println();
        Serial.println("[OTA] End");
    });
    ArduinoOTA.onError([](ota_error_t error) {
        Serial.printf("[OTA] Error[%u]: ", error);
             if (error == OTA_AUTH_ERROR)    { Serial.println("OTA_AUTH_ERROR"); }
        else if (error == OTA_BEGIN_ERROR)   { Serial.println("OTA_BEGIN_ERROR"); }
        else if (error == OTA_CONNECT_ERROR) { Serial.println("OTA_CONNECT_ERROR"); }
        else if (error == OTA_RECEIVE_ERROR) { Serial.println("OTA_RECEIVE_ERROR"); }
        else if (error == OTA_END_ERROR)     { Serial.println("OTA_END_ERROR"); }
    });
    
    ArduinoOTA.begin();
    
    Serial.println("[OTA] OTA is running");
}

/* WS2812
 * ========================================================================= */
#define WS2812_LEDS 120

#include "slc_register.h"
#define CLR_PERI_REG_MASK(reg, mask) CLEAR_PERI_REG_MASK(reg, mask)

typedef struct // type definition taken from : sdio_slv.h
{
    uint32_t blocksize:12;
    uint32_t datalen:12;
    uint32_t unused:5;
    uint32_t sub_sof:1;
    uint32_t eof:1;
    uint32_t owner:1;
    uint32_t buf_ptr;
    uint32_t next_link_ptr;

} sdio_queue_t;

// I2S timing parameters:
// Creates an I2S SR of 93,750 Hz, or 3 MHz Bitclock (.333us / sample)
// Measured on scope: 4 bitclock-ticks every 1200ns -> 300ns bitclock ???
// 12000000L / (div * bestbck * 2)
#define WS2812_I2S_BCK 17
#define WS2812_I2S_DIV 4

size_t   WS2812_bufferSize;
uint8_t *WS2812_buffer;

size_t   WS2812_i2sDataSize;
uint8_t *WS2812_i2sData;

size_t   WS2812_i2sZeroSize;
uint8_t *WS2812_i2sZero; 

sdio_queue_t WS2812_i2sDataQueue;
sdio_queue_t WS2812_i2sZeroQueue;

void WS2812_DMA()
{
    // Reset DMA
    SET_PERI_REG_MASK(SLC_CONF0, SLC_RXLINK_RST);
    CLR_PERI_REG_MASK(SLC_CONF0, SLC_RXLINK_RST);

    // Clear DMA int flags
    SET_PERI_REG_MASK(SLC_INT_CLR, 0xffffffff);
    CLR_PERI_REG_MASK(SLC_INT_CLR, 0xffffffff);

    // Enable and configure DMA
    CLR_PERI_REG_MASK(SLC_CONF0, (SLC_MODE << SLC_MODE_S));
    SET_PERI_REG_MASK(SLC_CONF0, (1        << SLC_MODE_S));
    
    SET_PERI_REG_MASK(SLC_RX_DSCR_CONF,
              SLC_INFOR_NO_REPLACE
            | SLC_TOKEN_NO_REPLACE);
    CLR_PERI_REG_MASK(SLC_RX_DSCR_CONF,
              SLC_RX_FILL_EN
            | SLC_RX_EOF_MODE
            | SLC_RX_FILL_MODE);

    // Configure DMA descriptor
    CLR_PERI_REG_MASK(SLC_RX_LINK,                                  SLC_RXLINK_DESCADDR_MASK);
    SET_PERI_REG_MASK(SLC_RX_LINK, ((uint32)&WS2812_i2sDataQueue) & SLC_RXLINK_DESCADDR_MASK);

    // Start transmission
    SET_PERI_REG_MASK(SLC_RX_LINK, SLC_RXLINK_START);

    // Init pins to i2s functions
    PIN_FUNC_SELECT(PERIPHS_IO_MUX_U0RXD_U, FUNC_I2SO_DATA);

    // Enable clock to i2s subsystem
    i2c_writeReg_Mask_def(i2c_bbpll, i2c_bbpll_en_audio_clock_out, 1);

    // Reset I2S subsystem
    CLR_PERI_REG_MASK(I2SCONF, I2S_I2S_RESET_MASK);
    SET_PERI_REG_MASK(I2SCONF, I2S_I2S_RESET_MASK);
    CLR_PERI_REG_MASK(I2SCONF, I2S_I2S_RESET_MASK);

    // Select 16bits per channel (FIFO_MOD=0), no DMA access (FIFO only)
    CLR_PERI_REG_MASK(I2S_FIFO_CONF,
               I2S_I2S_DSCR_EN
            | (I2S_I2S_RX_FIFO_MOD << I2S_I2S_RX_FIFO_MOD_S)
            | (I2S_I2S_TX_FIFO_MOD << I2S_I2S_TX_FIFO_MOD_S)
    );
    // Enable DMA in i2s subsystem
    SET_PERI_REG_MASK(I2S_FIFO_CONF, I2S_I2S_DSCR_EN);

    // trans master & rece slave, MSB shift, right_first, MSB right
    CLR_PERI_REG_MASK(I2SCONF,
               I2S_TRANS_SLAVE_MOD
            | (I2S_BITS_MOD << I2S_BITS_MOD_S)
            | (I2S_BCK_DIV_NUM << I2S_BCK_DIV_NUM_S)
            | (I2S_CLKM_DIV_NUM << I2S_CLKM_DIV_NUM_S)
    );
    SET_PERI_REG_MASK(I2SCONF,
              I2S_RIGHT_FIRST
            | I2S_MSB_RIGHT
            | I2S_RECE_SLAVE_MOD
            | I2S_RECE_MSB_SHIFT
            | I2S_TRANS_MSB_SHIFT
            | (((WS2812_I2S_BCK - 1) & I2S_BCK_DIV_NUM) << I2S_BCK_DIV_NUM_S)
            | (((WS2812_I2S_DIV - 1) & I2S_CLKM_DIV_NUM) << I2S_CLKM_DIV_NUM_S)
    );

    // Start transmission
    SET_PERI_REG_MASK(I2SCONF, I2S_I2S_TX_START);
}

// Init led-string / memory buffers etc.
void WS2812_Init(uint8_t leds, size_t bufferSize)
{
    WS2812_bufferSize  = bufferSize;
    WS2812_buffer      = (uint8_t *)calloc(WS2812_bufferSize,  1);
    
    WS2812_i2sDataSize = bufferSize * 4;
    WS2812_i2sData     = (uint8_t *)calloc(WS2812_i2sDataSize, 1); // buffer size * 4 i2s bits
    
    WS2812_i2sZeroSize = 28;
    WS2812_i2sZero     = (uint8_t *)calloc(WS2812_i2sZeroSize, 1); // WS2812 LED Treset = 67us (must be >50us)

    // set-up DMA descriptors

    WS2812_i2sDataQueue.blocksize     = WS2812_i2sDataSize;
    WS2812_i2sDataQueue.datalen       = WS2812_i2sDataSize;
    WS2812_i2sDataQueue.unused        = 0;
    WS2812_i2sDataQueue.sub_sof       = 0;
    WS2812_i2sDataQueue.eof           = 1;
    WS2812_i2sDataQueue.owner         = 1;
    WS2812_i2sDataQueue.buf_ptr       = (uint32_t) WS2812_i2sData;
    WS2812_i2sDataQueue.next_link_ptr = (uint32_t)&WS2812_i2sZeroQueue; // always link to zeros-buffer

    WS2812_i2sZeroQueue.blocksize     = WS2812_i2sZeroSize;
    WS2812_i2sZeroQueue.datalen       = WS2812_i2sZeroSize;
    WS2812_i2sZeroQueue.unused        = 0;
    WS2812_i2sZeroQueue.sub_sof       = 0;
    WS2812_i2sZeroQueue.owner         = 1;
    WS2812_i2sZeroQueue.eof           = 1;
    WS2812_i2sZeroQueue.buf_ptr       = (uint32_t) WS2812_i2sZero;
    WS2812_i2sZeroQueue.next_link_ptr = (uint32_t)&WS2812_i2sDataQueue; // link to DMA descriptor

    WS2812_DMA();
} 

// Send out WS2812 bits with coded pulses, one nibble, then the other.
static const uint16_t WS2812_BIT_PATTERNS[16] = {
    0b1000100010001000, 0b1000100010001110, 0b1000100011101000, 0b1000100011101110,
    0b1000111010001000, 0b1000111010001110, 0b1000111011101000, 0b1000111011101110,
    0b1110100010001000, 0b1110100010001110, 0b1110100011101000, 0b1110100011101110,
    0b1110111010001000, 0b1110111010001110, 0b1110111011101000, 0b1110111011101110,
};

void WS2812_Write(uint8_t *buffer, size_t length)
{
    // set-up pointers into the i2s-pixel-buffers
    uint16_t *i2s_ptr = (uint16_t *)WS2812_i2sData;
  
    while (length--) {
        uint8_t pixelByte = *buffer++;
        
        // transform pixel value into i2s nibble
        *(i2s_ptr++) = WS2812_BIT_PATTERNS[(pixelByte)      & 0x0F];
        *(i2s_ptr++) = WS2812_BIT_PATTERNS[(pixelByte >> 4) & 0x0F];
    }
}

/* WebSocket
 * ========================================================================= */
WebSocketsServer WebSockets = WebSocketsServer(80);

void WebSockets_Handle(uint8_t client, WStype_t type, uint8_t *payload, size_t length)
{
    switch(type) {
        case WStype_CONNECTED: {
            WebSockets.sendBIN(client, WS2812_i2sData, WS2812_i2sDataSize);
        } break;
        case WStype_BIN: {
            if (length > WS2812_bufferSize) {
                length = WS2812_bufferSize;
            }
            WS2812_Write(payload, length);
        } break;
    }
}

/* Pattern
 * ========================================================================= */
#define PATTERN_EANBLED 0

#if PATTERN_EANBLED != 0
float my_fmod(float arg1, float arg2) {
    int full = (int)(arg1/arg2);
    return arg1 - full*arg2;
}
uint64_t HSVtoHEX( float hue, float sat, float value )
{

    float pr,  pg, pb, avg;    pr=pg=pb=avg=0;
    short ora, og, ob;         ora=og=ob=0;

    float ro = my_fmod( hue * 6, 6. );
    ro = my_fmod( ro + 6 + 1, 6 ); //Hue was 60* off...

    //yellow->red
    if     ( ro < 1 ) { pr = 1;         pg = 1. - ro; }
    else if( ro < 2 ) { pr = 1;         pb = ro - 1.; }
    else if( ro < 3 ) { pr = 3. - ro;   pb = 1;       }
    else if( ro < 4 ) { pb = 1;         pg = ro - 3;  }
    else if( ro < 5 ) { pb = 5  - ro;   pg = 1;       }
    else              { pg = 1;         pr = ro - 5;  }

    //Actually, above math is backwards, oops!
    pr *= value;   pg *= value;   pb *= value;
    avg += pr;     avg += pg;     avg += pb;

    pr = pr * sat + avg * (1.-sat);
    pg = pg * sat + avg * (1.-sat);
    pb = pb * sat + avg * (1.-sat);

    ora = pr*255;  og = pb*255;   ob = pg*255;

    if( ora < 0 ) ora = 0;   if( ora > 255 ) ora = 255;
    if( og  < 0 ) og = 0;    if( og  > 255 ) og  = 255;
    if( ob  < 0 ) ob = 0;    if( ob > 255 )  ob  = 255;

    return (ob<<16) | (og<<8) | ora;
}
const unsigned char rands[] = {
	0x67, 0xc6, 0x69, 0x73, 0x51, 0xff, 0x4a, 0xec, 0x29, 0xcd, 0xba, 0xab, 0xf2, 0xfb, 0xe3, 0x46, 
	0x7c, 0xc2, 0x54, 0xf8, 0x1b, 0xe8, 0xe7, 0x8d, 0x76, 0x5a, 0x2e, 0x63, 0x33, 0x9f, 0xc9, 0x9a, 
	0x66, 0x32, 0x0d, 0xb7, 0x31, 0x58, 0xa3, 0x5a, 0x25, 0x5d, 0x05, 0x17, 0x58, 0xe9, 0x5e, 0xd4, 
	0xab, 0xb2, 0xcd, 0xc6, 0x9b, 0xb4, 0x54, 0x11, 0x0e, 0x82, 0x74, 0x41, 0x21, 0x3d, 0xdc, 0x87, 
	0x70, 0xe9, 0x3e, 0xa1, 0x41, 0xe1, 0xfc, 0x67, 0x3e, 0x01, 0x7e, 0x97, 0xea, 0xdc, 0x6b, 0x96, 
	0x8f, 0x38, 0x5c, 0x2a, 0xec, 0xb0, 0x3b, 0xfb, 0x32, 0xaf, 0x3c, 0x54, 0xec, 0x18, 0xdb, 0x5c, 
	0x02, 0x1a, 0xfe, 0x43, 0xfb, 0xfa, 0xaa, 0x3a, 0xfb, 0x29, 0xd1, 0xe6, 0x05, 0x3c, 0x7c, 0x94, 
	0x75, 0xd8, 0xbe, 0x61, 0x89, 0xf9, 0x5c, 0xbb, 0xa8, 0x99, 0x0f, 0x95, 0xb1, 0xeb, 0xf1, 0xb3, 
	0x05, 0xef, 0xf7, 0x00, 0xe9, 0xa1, 0x3a, 0xe5, 0xca, 0x0b, 0xcb, 0xd0, 0x48, 0x47, 0x64, 0xbd, 
	0x1f, 0x23, 0x1e, 0xa8, 0x1c, 0x7b, 0x64, 0xc5, 0x14, 0x73, 0x5a, 0xc5, 0x5e, 0x4b, 0x79, 0x63, 
	0x3b, 0x70, 0x64, 0x24, 0x11, 0x9e, 0x09, 0xdc, 0xaa, 0xd4, 0xac, 0xf2, 0x1b, 0x10, 0xaf, 0x3b, 
	0x33, 0xcd, 0xe3, 0x50, 0x48, 0x47, 0x15, 0x5c, 0xbb, 0x6f, 0x22, 0x19, 0xba, 0x9b, 0x7d, 0xf5, 
	0x0b, 0xe1, 0x1a, 0x1c, 0x7f, 0x23, 0xf8, 0x29, 0xf8, 0xa4, 0x1b, 0x13, 0xb5, 0xca, 0x4e, 0xe8, 
	0x98, 0x32, 0x38, 0xe0, 0x79, 0x4d, 0x3d, 0x34, 0xbc, 0x5f, 0x4e, 0x77, 0xfa, 0xcb, 0x6c, 0x05, 
	0xac, 0x86, 0x21, 0x2b, 0xaa, 0x1a, 0x55, 0xa2, 0xbe, 0x70, 0xb5, 0x73, 0x3b, 0x04, 0x5c, 0xd3, 
	0x36, 0x94, 0xb3, 0xaf, 0xe2, 0xf0, 0xe4, 0x9e, 0x4f, 0x32, 0x15, 0x49, 0xfd, 0x82, 0x4e, 0xa9, };
const unsigned char sintable[] = {
	0x80, 0x83, 0x86, 0x89, 0x8c, 0x8f, 0x92, 0x95, 0x99, 0x9c, 0x9f, 0xa2, 0xa5, 0xa8, 0xab, 0xad, 
	0xb0, 0xb3, 0xb6, 0xb9, 0xbc, 0xbe, 0xc1, 0xc4, 0xc6, 0xc9, 0xcb, 0xce, 0xd0, 0xd3, 0xd5, 0xd7, 
	0xda, 0xdc, 0xde, 0xe0, 0xe2, 0xe4, 0xe6, 0xe8, 0xe9, 0xeb, 0xed, 0xee, 0xf0, 0xf1, 0xf3, 0xf4, 
	0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfc, 0xfd, 0xfe, 0xfe, 0xfe, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xfe, 0xfe, 0xfe, 0xfd, 0xfc, 0xfc, 0xfb, 0xfa, 0xf9, 0xf8, 0xf7, 0xf6, 
	0xf5, 0xf4, 0xf3, 0xf1, 0xf0, 0xee, 0xed, 0xeb, 0xe9, 0xe8, 0xe6, 0xe4, 0xe2, 0xe0, 0xde, 0xdc, 
	0xda, 0xd7, 0xd5, 0xd3, 0xd0, 0xce, 0xcb, 0xc9, 0xc6, 0xc4, 0xc1, 0xbe, 0xbc, 0xb9, 0xb6, 0xb3, 
	0xb0, 0xad, 0xab, 0xa8, 0xa5, 0xa2, 0x9f, 0x9c, 0x99, 0x95, 0x92, 0x8f, 0x8c, 0x89, 0x86, 0x83, 
	0x80, 0x7d, 0x79, 0x76, 0x73, 0x70, 0x6d, 0x6a, 0x67, 0x64, 0x61, 0x5e, 0x5b, 0x58, 0x55, 0x52, 
	0x4f, 0x4c, 0x49, 0x47, 0x44, 0x41, 0x3e, 0x3c, 0x39, 0x36, 0x34, 0x31, 0x2f, 0x2d, 0x2a, 0x28, 
	0x26, 0x24, 0x21, 0x1f, 0x1d, 0x1b, 0x1a, 0x18, 0x16, 0x14, 0x13, 0x11, 0x10, 0x0e, 0x0d, 0x0b, 
	0x0a, 0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x04, 0x03, 0x02, 0x02, 0x01, 0x01, 0x01, 0x01, 0x01, 
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x02, 0x02, 0x03, 0x04, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 
	0x0a, 0x0b, 0x0d, 0x0e, 0x10, 0x11, 0x13, 0x14, 0x16, 0x18, 0x1a, 0x1b, 0x1d, 0x1f, 0x21, 0x24, 
	0x26, 0x28, 0x2a, 0x2d, 0x2f, 0x31, 0x34, 0x36, 0x39, 0x3c, 0x3e, 0x41, 0x44, 0x47, 0x49, 0x4c, 
	0x4f, 0x52, 0x55, 0x58, 0x5b, 0x5e, 0x61, 0x64, 0x67, 0x6a, 0x6d, 0x70, 0x73, 0x76, 0x79, 0x7d, };
unsigned char intsin( short s )
{
	return sintable[((unsigned short)s)>>8];
}
short dtr[512];
int init_dtr = 0;
uint32_t hex_pattern( uint8_t pattern, uint16_t light, uint16_t lights, uint32_t frame) {
    uint32_t hex = 0; //0xbbggrr
	int i;

    switch( pattern ) {
        // for button
        case 1:
            hex = (light == (frame % lights)) ? 0xFFFFFF : 0x000000;
            break;
        case 2:
            hex = HSVtoHEX(light * (1/12.) + frame * (1./48.), 1, 1.0);
            break;
        case 3:
            hex = ((int32_t)((frame+light)%lights)>(lights-2)) ? 0xffffff : 0;
            break; // the worm.
        case 4:
            hex = HSVtoHEX(light * .03 - frame * .04, 1, (((((-light << 3) % 256) - ((frame << 3) % 256) + 256) % 256)) /256.0 * 0.9 - 0.1);
            break; // long, smooth, transitioning. 1.0

        // for wall art
        case 5:
            hex = (((frame + light) % 186) > 160) ? 0xff8f8f : 0;
            break; // the worm.
        
        case 6:
            hex = (((frame + light) % 186) > 130) ? 0x0000ff : 0;
            break; // the red worm.
        
        case 7:
            hex = HSVtoHEX(light*.005 - frame*.001, 1, ((((-light%256) - ((frame>>1)%256)+256)%256) ) /256.0*1.2-0.1);
            break;
        
        case 8:
            hex = HSVtoHEX(light*.00500 + ((int)(frame*0.42))*.17, 1, 0.40 );
            break;// fiesta

        // and my favorite:
        case 9:
            hex = HSVtoHEX(light*.001 - frame*.001, 1,  ((((light%256) - ((frame>>1)%256)+256)%256) ) /256.0*1.5-0.1);
            break; // long, smooth, transitioning. 1.0 overloads.  Trying 0.9. If 0.9 works, should back off.

        case 10:
            hex = HSVtoHEX(light*0.005376344 - frame*.001, 1.3,  1.0);
            break; // long, smooth, transitioning. full-brigth

        // chaser
        case 12:
            hex = HSVtoHEX(light*.002 + frame*.002, 4, 1.0);
            break; // 0.50 = overload. 0.45 = overheat? =0.40 = HOT

        case 13:
            hex = HSVtoHEX(frame*.001, 1.0,  1.0);
            break; // long, smooth, transitioning. full-brigth

        case 14: {
            float v1 = (((light%256) << 2) - ((frame    ) % 1024) + 1024) % 1024;
            float v2 = (((light%256) << 2) - ((frame + 1) % 1024) + 1024) % 1024;
            float v3 = (((light%256) << 2) - ((frame + 2) % 1024) + 1024) % 1024;
            float v4 = (((light%256) << 2) - ((frame + 3) % 1024) + 1024) % 1024;
            hex = HSVtoHEX(light * .001 - frame * .001, 1, (v1 + v2 + v3 + v4) / (4.0 * 1024.0) * 1.2 - 0.1); 
            break; // long, smooth, transitioning. 1.0 overloads.  Trying 0.9. If 0.9 works, should back off.
        }
        
        case 15: {
            int li = light/2;
            if (!init_dtr) {
                for(i = 0; i < (sizeof(dtr) / sizeof(dtr[0])); i++) {
                    dtr[i] = i*10; init_dtr = 1;
                }
            }
            dtr[li] += rands[(li + (frame >> 6)) % (sizeof(rands) / sizeof(rands[0]))];
            hex = (int)(((short)intsin(frame * 2 + dtr[li] * 4)) * 1.0);
            if (hex > 0xff) {
                hex = 0xff;
            }
            hex = (0xff << 16) | (hex) | (hex << 8);
            break;
        }
 
        default: hex = HSVtoHEX(light * .05 + frame * .01, 1, 1.0);
    }

    return hex;
}
#endif

void setup()
{
    pinMode(LED_BUILTIN, OUTPUT); // Enable built-in LED
    
    Serial.begin(115200);
    do { delay(100); } while (!Serial); // wait for serial connection
    
    Serial.println();
    Serial.println("WS2812 " __DATE__ " " __TIME__);
    Serial.println("Booting...");

    ESP_Boot_WiFiConnect(ESP_WIFI_SSID, ESP_WIFI_PSK);
    ESP_Boot_EnableOTA(ESP_OTA_HOSTNANE, ESP_OTA_PASSWORD);
    
    Serial.println("Boot OK!");

    WS2812_Init(WS2812_LEDS, WS2812_LEDS * 3); // 120 leds * 3 (RGB bits)
    Serial.println("[WS2812 Driver] OK");
    
    WebSockets.begin();
    WebSockets.onEvent(WebSockets_Handle);
    Serial.println("[WebSockets Driver] OK");
}

void loop()
{
#if PATTERN_EANBLED != 0
    static uint32_t t = 0;
    t++;
    for (uint16_t i = 0, l = WS2812_LEDS; i < l; i++) {
        uint32_t c = hex_pattern(9, i, l, t);
        
        uint8_t *a =   (uint8_t *)&c;
        uint8_t *b = &((uint8_t *)WS2812_buffer)[i * 3];

        //memcpy(b, a, 3);
        b[0] = a[0];
        b[1] = a[1];
        b[2] = a[2];
    }
    WS2812_Write(WS2812_buffer, WS2812_bufferSize);
#endif

    delay(9);

    WebSockets.loop();
    ArduinoOTA.handle();

    delay(1);
}
